/* op_gen.c  <9-30-90>  <3-24-92> TGR */

 /***************************************************************************** 
 *
 * $Id$
 *
 * $Log:	op_gen.c,v $
 * Revision 1.4  93/03/10  11:17:40  11:17:40  rossin (Ted Rossin)
 * Added code to get_const to remove trailing white space from input string.
 * 
 * Revision 1.3  93/03/10  11:06:45  11:06:45  rossin (Ted Rossin)
 * Fixed bug of not dealing with spaces and tabs before and after parameters.
 * 
 * Revision 1.2  93/03/05  13:26:18  13:26:18  rossin (Ted Rossin)
 * Fixed symbol truncation bug that caused an error instead of a warning.
 * Also fixed warning messages in assembler listing output.
 * 
 * 
 * Revision 1.1  93/01/09  11:05:38  11:05:38  rossin (Ted Rossin)
 * Initial revision
 * 
 *****************************************************************************/

#include "port.h"
#include <ctype.h>
#ifdef MAC
#include <string.h> 
#endif
#include "as30.h"
#include "ophash.h"

extern char *sym,*op,*parms;
extern char *sym2,*op2,*parms2;
extern int sym_ptr;

extern char *global_list[MAX_SYM];
extern Sym_tab *sym_tab;
extern Image *image;
extern Sec_loc_cnt *sec_loc_cnt;
extern Rel_tab *rel_tab;

extern Op_table op_table[];

extern int current_section;
extern long image_count;
extern Image *img_ptr;
extern int rel_tab_ptr;
extern int fp_stack_pointer;

#define INTERLOCKED	0x80
#define FLAG		0x40
#define FLOAT_SOURCE	0x20
#define FLOAT_DEST	0x10

#define MAX_OP_PARMS	7
#define REGISTER	0
#define DIRECT		1
#define INDIRECT	2
#define IMMEDIATE	3

#define NUM_IND_MODES	26
#define NUM_REGS	28

#define NON_DELAYED	0
#define DELAYED		1

gen_op(line_num,op_code,parameters,error)
unsigned int line_num;
char *op_code,*parameters,*error;
{
    int i,start,stop,return_val;

    if(op_code[0] == '.'){
	return(decode_pseudo_op(line_num,op_code,parameters,error));
    }
    if(op_code[0]<'a' || op_code[0]>'z'){
	img_ptr->file_number=fp_stack_pointer;
	img_ptr->line_num=line_num;  
	img_ptr->section=current_section;  
	img_ptr->data=0L;  
    	img_ptr->address=sec_loc_cnt[current_section].lc;
    	sec_loc_cnt[current_section].lc++;
	inc_image_pointer();
	sprintf(error,"Invalid opcode %s",op_code);
	return(-1);
    }

    start=op_code_hash[op_code[0]-'a'];
    stop=op_code_last[op_code[0]-'a'];
    for(i=start;i<stop;i++){
	if(!strcmp(op_code,op_table[i].mnemonic)){ 
	    img_ptr->file_number=fp_stack_pointer;
	    img_ptr->line_num=line_num;  
	    img_ptr->section=current_section;  
	    return_val=encode_parms(line_num,parameters,op_table[i].operation
		,&(img_ptr->data),i,error);
	    if(return_val) return(return_val);
    	    img_ptr->address=sec_loc_cnt[current_section].lc;
    	    sec_loc_cnt[current_section].lc++;
	    inc_image_pointer();
	    return(return_val);
	}
    }
    img_ptr->file_number=fp_stack_pointer;
    img_ptr->line_num=line_num;  
    img_ptr->section=current_section;  
    img_ptr->data=0L;  
    img_ptr->address=sec_loc_cnt[current_section].lc;
    sec_loc_cnt[current_section].lc++;
    inc_image_pointer();
    sprintf(error,"Invalid opcode %s",op_code);
    return(-1);
}

encode_parms(line_num,parm,op_code,instruction,op_index,error)
char *parm,*error;
unsigned long op_code,*instruction;
int op_index;
{
    int parm_num=0,return_val=0,group,special_number,disp_24,itype;
    unsigned long work_code;
    char expr[100],*ptr=expr;

    static char symbol[MAX_OP_PARMS][MAX_SYM];
    static unsigned long type[MAX_OP_PARMS],reg_num[MAX_OP_PARMS];
    static unsigned long mod[MAX_OP_PARMS];
    static unsigned long disp[MAX_OP_PARMS];

    work_code=0;
    group=op_table[op_index].group;
    special_number = op_table[op_index].special_number;

    disp_24=(group==4 || (group==6 && special_number==LDP))?1:0;
    itype=(group&FLOAT_SOURCE)?1:0;
    for( ; *parm; parm++){
	if(*parm != ','){
	    if(!isspace(*parm)) *ptr++ = *parm;
    	}
    	else{
	    *ptr=0;
	    ptr=expr;
	    if(decode_addr_mode(disp_24,itype,expr,&type[parm_num]
		,&reg_num[parm_num]
		,&mod[parm_num],&disp[parm_num],symbol[parm_num],error)){
		return_val = -1;  break;
	    }
	    parm_num++;
	    if(parm_num>=MAX_OP_PARMS){
		sprintf(error,"Too many parameters");
	   	return_val = -1;  break;
	    }
    	}
    }
    *ptr=0;
    if(expr[0]){
	if(decode_addr_mode(disp_24,itype,expr,&type[parm_num]
	    ,&reg_num[parm_num]
	    ,&mod[parm_num],&disp[parm_num],symbol[parm_num],error)){
	    return_val = -1;
	}
	parm_num++;
	if(parm_num>=MAX_OP_PARMS){
	    sprintf(error,"Too many parameters");
	    return_val = -1;
	}
    }
    if(return_val){
	img_ptr->file_number=fp_stack_pointer;
    	img_ptr->line_num=line_num;  
    	img_ptr->section=current_section;  
    	img_ptr->data=0L;  
    	img_ptr->address=sec_loc_cnt[current_section].lc;
    	sec_loc_cnt[current_section].lc++;
    	inc_image_pointer();
 	return(-1);
    }

	/* construct op code */

    switch(group & 0x0F){	 
	case 0:	/* General Load(2 operands) */
	    return_val=general_load(parm_num,type,reg_num,mod,disp,symbol
						,group,error,&work_code);
	    break;
	case 1:	/* 3 Operand (2,3 operands) */
	    return_val=three_op(parm_num,type,reg_num,mod,disp
						,group,error,&work_code);
	    break;
	case 2:	/* Parallel 6 (6 operands) */
	    return_val=six_op(parm_num,type,reg_num,mod,disp,error,&work_code);
	    break;
	case 3:	/* Parallel 4/5 (4,5 operands) */
	    return_val=four_five_op(parm_num,type,reg_num,mod,disp
				,special_number,error,&work_code);
	    break;
	case 4:	/* Long immediate (1 operand) */
	    return_val=long_immediate(parm_num,type,disp,symbol,error
								,&work_code);
	    break;
	case 5:	/* Non delayed cond branch or call (1 or 2 operands) */
	    return_val=branch_or_call(NON_DELAYED,parm_num,type,reg_num
					,disp,symbol,group,error,&work_code);
	    break;
	case 6:	/* Special (? operands uses special field) */
	    return_val=special_instruct(parm_num,type,reg_num,mod,disp
				,symbol,group,special_number,error,&work_code);
	    break; 
	case 7:	/* No operands (0 operands) */
	    break;
	case 8:	/* General Store (2 operands) */
	    return_val=general_store(parm_num,type,reg_num,mod,disp,symbol
						,group,error,&work_code);
	    break;
	case 9:	/* Trap instruction (1 operand) */
	    return_val=trap_op(parm_num,type,disp,error,&work_code);
	    break;
	case 10: /* Delayed Conditional Branch (1 or 2 operands) */
	    return_val=branch_or_call(DELAYED,parm_num,type,reg_num
					,disp,symbol,group,error,&work_code);
	    break;
	default:
	    sprintf(error,"Invalid group code in opcode table");
	    return_val = -1;
	    break;
    }
    if(return_val){
	img_ptr->file_number=fp_stack_pointer;
    	img_ptr->line_num=line_num;  
    	img_ptr->section=current_section;  
    	img_ptr->data=0L;  
    	img_ptr->address=sec_loc_cnt[current_section].lc;
    	sec_loc_cnt[current_section].lc++;
    	inc_image_pointer();
 	return(-1);
    }
    *instruction = op_code | work_code;
    return(0);
}

six_op(parm_num,type,reg_num,mod,disp,error,parm_field)
int parm_num;
unsigned long type[],reg_num[],mod[],disp[],*parm_field;
char *error;
{
    unsigned long work_code=0,p_field;
    int i,line,s1,s2,s3,s4;

    *parm_field=0;
    if(parm_num != 6){
    	sprintf(error,"Instruction requires 6 operands");
    	return(-1);
    }
    if(type[2]!=REGISTER || reg_num[2]>1){
	strcpy(error,"Operand 3 of first instruction must be r0 or r1");
	return(-1);
    }
    if(type[5]!=REGISTER || reg_num[5]>3 || reg_num[5]<2){
	strcpy(error,"Operand 3 of second instruction must be r0 or r1");
	return(-1);
    }
    reg_num[5] &= 1L;  /* only need low bit */

    if(	    type[4]==REGISTER && type[3]==REGISTER){
	    p_field=0;
	    if(type[0]==INDIRECT && type[1]==INDIRECT){
	    	s1=4; s2=3; s3=0; s4=1;
	    }
	    else p_field=0xFFFF; /* Won't fit */
    }
    else if(type[4]==INDIRECT && type[3]==REGISTER){
	    p_field=1;
	    if(type[0]==INDIRECT && type[1]==REGISTER){
	    	s1=1; s2=3; s3=0; s4=4;
	    }
	    else{
		if(type[0]==REGISTER && type[1]==INDIRECT){
	    	    s1=0; s2=3; s3=1; s4=4;  /* flip mult operands */
	        }
		else p_field=0xFFFF;  /* Won't fit */
	    }
    }
    else if(type[4]==INDIRECT && type[3]==INDIRECT){
	    p_field=2;
	    if(type[0]==REGISTER && type[1]==REGISTER){
	    	s1=0; s2=1; s3=4; s4=3;
	    }
	    else p_field=0xFFFF; /* Won't fit */
    }
    else if(type[4]==REGISTER && type[3]==INDIRECT){
	    p_field=3;
	    if(type[0]==INDIRECT && type[1]==REGISTER){
	    	s1=1; s2=4; s3=0; s4=3;
	    }
	    else{
		if(type[0]==REGISTER && type[1]==INDIRECT){
	    	    s1=0; s2=4; s3=1; s4=3; /* flip mult operands */
	        }
		else p_field=0xFFFF;  /* Won't fit */
	    }
    }
    if(p_field==0xFFFF){
	strcpy(error,"Source operands must consist of 2 register ");
 	strcat(error,"and 2 indirect modes");
	return(-1);
    }
    for(line=1;line<2;line++){
    	for(i=1;i<3;i++){	
	    if(reg_num[3*(line-1)+i-1]>7){
		sprintf(error,"operand %d of instuction %d: register ",i,line);
		strcat(error,"must be extended precision register");
		return(-1);
	    }
	    if(disp[3*(line-1)+i-1]>1){
		sprintf(error,"operand %d of instuction %d: ",i,line);
		strcat(error,"indirect must have disp=0,1,ir0,ir1");
		return(-1);
	    }
	}
    }
    work_code=( (p_field<<24)|(reg_num[2]<<23)|(reg_num[5]<<22) );
    if(mod[s3]<8 && disp[s3]==0) /* Change addressing mode to no disp */
	mod[s3]=0x18;
    if(mod[s4]<8 && disp[s4]==0) /* Change addressing mode to no disp */
	mod[s4]=0x18;
    work_code |= ( (reg_num[s1]<<19)|(reg_num[s2]<<16));
    work_code |= ( (mod[s3]<<11)|(reg_num[s3]<<8));
    work_code |= ( (mod[s4]<<3)|(reg_num[s4]));
    *parm_field=work_code;
    return(0);
}

four_five_op(parm_num,type,reg_num,mod,disp,special,error,parm_field)
int parm_num,special;
unsigned long type[],reg_num[],mod[],disp[],*parm_field;
char *error;
{
    unsigned long work_code=0L;
    int required_parms;

    *parm_field=0;
    required_parms=(special==1)?5:4;	/* If flag set no destination */
    if(parm_num != required_parms){
    	sprintf(error,"Instruction requires %d operands",required_parms);
    	return(-1);
    }
    switch(special){
	case 0: /* d1 00 s3 d2 s2  op(s2)->d1  || s3->d2 (example absf||stf) */
	    if(type[0]!=INDIRECT || disp[0]>1){
	       	strcpy(error,"Operand 1 of first instruction must be indirect");
	       	strcat(error," (with disp=0,1,ir0,ir1)");
               	return(-1);
	    }
	    if(type[1]!=REGISTER || reg_num[1]>7){
		strcpy(error,"Operand 2 of first instruction must be extended");
		strcat(error," precision register");
		return(-1);
	    }
	    if(type[2]!=REGISTER || reg_num[2]>7){
		strcpy(error,"Operand 1 of 2nd instruction must be extended");
		strcat(error," precision register");
		return(-1);
	    }
	    if(type[3]!=INDIRECT || disp[3]>1){
	       	strcpy(error,"Operand 2 of 2nd instruction must be indirect");
	       	strcat(error," (with disp=0,1,ir0,ir1)");
               	return(-1);
	    }
	    if(mod[0]<8 && disp[0]==0) /* Change addressing mode to no disp */
		mod[0]=0x18;
	    if(mod[3]<8 && disp[3]==0) /* Change addressing mode to no disp */
		mod[3]=0x18;
	    work_code = ((reg_num[1]<<22) | (reg_num[2]<<16));
	    work_code |= ((mod[3]<<11)|(reg_num[3]<<8));	    
	    work_code |= ((mod[0]<<3)|(reg_num[0]));	    
	    break;
	case 1: /* d1 s1 s3 d2 s2  s1 op s2 ->d1  || s3->d2 (ex addf3||stf) */
	    if(type[0]!=INDIRECT || disp[0]>1){
	       	strcpy(error,"Operand 1 of first instruction must be indirect");
	       	strcat(error," (with disp=0,1,ir0,ir1)");
               	return(-1);
	    }
	    if(type[1]!=REGISTER || reg_num[1]>7){
		strcpy(error,"Operand 2 of first instruction must be extended");
		strcat(error," precision register");
		return(-1);
	    }
	    if(type[2]!=REGISTER || reg_num[2]>7){
		strcpy(error,"Operand 3 of first instruction must be extended");
		strcat(error," precision register");
		return(-1);
	    }
	    if(type[3]!=REGISTER || reg_num[3]>7){
		strcpy(error,"Operand 1 of 2nd instruction must be extended");
		strcat(error," precision register");
		return(-1);
	    }
	    if(type[4]!=INDIRECT || disp[4]>1){
	       	strcpy(error,"Operand 2 of 2nd instruction must be indirect");
	       	strcat(error," (with disp=0,1,ir0,ir1)");
               	return(-1);
	    }
	    if(mod[0]<8 && disp[0]==0) /* Change addressing mode to no disp */
		mod[0]=0x18;
	    if(mod[4]<8 && disp[4]==0) /* Change addressing mode to no disp */
		mod[4]=0x18;
	    work_code = ((reg_num[2]<<22)|(reg_num[1]<<19)|(reg_num[3]<<16));
	    work_code |= ((mod[4]<<11)|(reg_num[4]<<8));	    
	    work_code |= ((mod[0]<<3)|(reg_num[0]));	    
	    break;
	case 2: /* d2 d1 00 s1 s2  s2->d2  || s1->d1  (example ldf||ldf) */
	    if(type[0]!=INDIRECT || disp[0]>1){
	       	strcpy(error,"Operand 1 of first instruction must be indirect");
	       	strcat(error," (with disp=0,1,ir0,ir1)");
               	return(-1);
	    }
	    if(type[1]!=REGISTER || reg_num[1]>7){
		strcpy(error,"Operand 2 of first instruction must be extended");
		strcat(error," precision register");
		return(-1);
	    }
	    if(type[2]!=INDIRECT || disp[2]>1){
	       	strcpy(error,"Operand 1 of 2nd instruction must be indirect");
	       	strcat(error," (with disp=0,1,ir0,ir1)");
               	return(-1);
	    }
	    if(type[3]!=REGISTER || reg_num[3]>7){
		strcpy(error,"Operand 2 of 2nd instruction must be extended");
		strcat(error," precision register");
		return(-1);
	    }
	    if(mod[0]<8 && disp[0]==0) /* Change addressing mode to no disp */
		mod[0]=0x18;
	    if(mod[2]<8 && disp[2]==0) /* Change addressing mode to no disp */
		mod[2]=0x18;
	    work_code = ((reg_num[1]<<22) | (reg_num[3]<<19));
	    work_code |= ((mod[2]<<11)|(reg_num[2]<<8));	    
	    work_code |= ((mod[0]<<3)|(reg_num[0]));	    
	    break;
	case 3: /* s2 00 s1 d1 d2  s2->d2  || s1->d1 (example stf||stf) */
	    if(type[0]!=REGISTER || reg_num[0]>7){
		strcpy(error,"Operand 1 of first instruction must be extended");
		strcat(error," precision register");
		return(-1);
	    }
	    if(type[1]!=INDIRECT || disp[1]>1){
	       	strcpy(error,"Operand 2 of first instruction must be indirect");
	       	strcat(error," (with disp=0,1,ir0,ir1)");
               	return(-1);
	    }
	    if(type[2]!=REGISTER || reg_num[2]>7){
		strcpy(error,"Operand 1 of 2nd instruction must be extended");
		strcat(error," precision register");
		return(-1);
	    }
	    if(type[3]!=INDIRECT || disp[3]>1){
	       	strcpy(error,"Operand 2 of 2nd instruction must be indirect");
	       	strcat(error," (with disp=0,1,ir0,ir1)");
               	return(-1);
	    }
	    if(mod[1]<8 && disp[1]==0) /* Change addressing mode to no disp */
		mod[1]=0x18;
	    if(mod[3]<8 && disp[3]==0) /* Change addressing mode to no disp */
		mod[3]=0x18;
	    work_code = ((reg_num[0]<<22) | (reg_num[2]<<16));
	    work_code |= ((mod[3]<<11)|(reg_num[3]<<8));	    
	    work_code |= ((mod[1]<<3)|(reg_num[1]));	    
	    break;
	default:
  	    sprintf(error,"special field invalid in op-code table");
    	    return(-1);
	    break;
    }
    *parm_field=work_code;
    return(0);
}

branch_or_call(delay,parm_num,type,reg_num,disp,symbol,group,error,parm_field)
int delay,parm_num,group;
unsigned long type[],disp[],reg_num[],*parm_field;
char symbol[][MAX_SYM],*error;
{
    unsigned long work_code=0;
    int num_parms;

    *parm_field=0;
    num_parms=(group&FLAG)?2:1;
    if(parm_num != num_parms){
	sprintf(error,"Instruction requires %d operand(s)",num_parms);
	return(-1);
    }
    if(num_parms==2 && (type[0]!=REGISTER || reg_num[0]<8 || reg_num[0]>0x0F)){
	sprintf(error,"source 1 must be auxiliary register (ar0->ar7)");
	return(-1);
    }
    if(num_parms == 2){ 
	work_code = ((reg_num[0]&7)<<22);
	if(type[1] == IMMEDIATE) work_code |= (1L<<25);
    }
    else	
    	work_code = (type[0]==IMMEDIATE)?(1L<<25):0;
    switch((int)type[num_parms-1]){
	 case REGISTER:
	    work_code |= reg_num[num_parms-1];
	    break;
      	case IMMEDIATE:
	    if(symbol[num_parms-1][0]){ /* Symbol */
	        rel_tab[rel_tab_ptr].section=current_section;
	        rel_tab[rel_tab_ptr].offset=sec_loc_cnt[current_section].lc;
	        rel_tab[rel_tab_ptr].image_offset=image_count;
	        rel_tab[rel_tab_ptr].flags=(delay)?R_PCRDWORD:R_PCRWORD;
	        strncpy(rel_tab[rel_tab_ptr].symbol,symbol[num_parms-1]
					,MAX_SYM-1);
		rel_tab[rel_tab_ptr].symbol[MAX_SYM-1]=0;
	        inc_rel_pointer();
	    }
	    else{
	  	work_code |= disp[num_parms-1];
	    }
	    break;
    }

    *parm_field=work_code;
    return(0);
}

special_instruct(parm_num,type,reg_num,mod,disp,symbol,group,special_number
	,error,parm_field)
int parm_num,group,special_number;
unsigned long type[],disp[],reg_num[],mod[],*parm_field;
char symbol[][MAX_SYM],*error;
{
    unsigned long work_code=0;

    *parm_field=0;
    switch(special_number){
	case IACK:
	    if(parm_num != 1){
		sprintf(error,"Instruction requires one operand");
		return(-1);
	    }
	    work_code = (type[0]<<21);
    	    switch((int)type[0]){
	      case DIRECT:
	    	if(symbol[0][0]){ /* Symbol */
	    	  rel_tab[rel_tab_ptr].section=current_section;
	     	  rel_tab[rel_tab_ptr].offset=sec_loc_cnt[current_section].lc;
	    	  rel_tab[rel_tab_ptr].image_offset=image_count;
	    	  rel_tab[rel_tab_ptr].flags=R_RELWORD;
	    	  strncpy(rel_tab[rel_tab_ptr].symbol,symbol[0],MAX_SYM-1);
		  rel_tab[rel_tab_ptr].symbol[MAX_SYM-1]=0;
	    	  inc_rel_pointer();
	    	}
	    	else{
		  work_code |= disp[0];
	    	}
	    	break;
	      case INDIRECT:
	    	if(disp[0]>0x00FF){ 
	    	  sprintf(error,
		    "Displacement must fit into an unsigned 8-bit field");
	    	  return(-1);
	    	}
	    	work_code |= ((mod[0]<<11) | (reg_num[0]<<8) | disp[0]);
	    	break;
	      default:
	 	sprintf(error,"Source must be DIRECT or INDIRECT mode");
	 	return(-1);
		break;
    	    }
	    break;

	case NOP:
	   
	    if(parm_num == 0) return(0);  /* Set parm field to zero */
	    if(parm_num != 1){
	 	sprintf(error,"Instruction requires 1 or no parmeters");
	 	return(-1);
	    }
	    work_code = (type[0]<<21);
    	    switch((int)type[0]){
	      case INDIRECT:
	    	if(disp[0]>0x00FF){ 
	    	  sprintf(error,
		    "Displacement must fit into an unsigned 8-bit field");
	    	  return(-1);
	    	}
	    	work_code |= ((mod[0]<<11) | (reg_num[0]<<8) | disp[0]);
	    	break;
	      default:
	 	sprintf(error,"Source must be INDIRECT if specified");
	 	return(-1);
		break;
	    }
	    break;

	case POP:
	case POPF:
	case PUSH:
	case PUSHF:
	case ROL:
	case ROLC:
	case ROR:
	case RORC:
	    if(parm_num != 1 || type[0] != REGISTER){
		sprintf(error,"Instruction requires 1 register parameter");
		return(-1);
	    }
	    if( (group&FLOAT_SOURCE || group&FLOAT_DEST) && reg_num[0]>7){
		sprintf(error,"Extended precision register required");
		return(-1);
	    }
	    work_code = (reg_num[0]<<16);
	    break;

	case RPTS:

	    if(parm_num != 1){
		sprintf(error,"Instruction requires one operand");
		return(-1);
	    }
	    work_code = (type[0]<<21);
    	    switch((int)type[0]){
	      case REGISTER:
	    	work_code |= reg_num[0];
	    	break;
	      case DIRECT:
	    	if(symbol[0][0]){ /* Symbol */
	    	  rel_tab[rel_tab_ptr].section=current_section;
	     	  rel_tab[rel_tab_ptr].offset=sec_loc_cnt[current_section].lc;
	    	  rel_tab[rel_tab_ptr].image_offset=image_count;
	    	  rel_tab[rel_tab_ptr].flags=R_RELWORD;
	    	  strncpy(rel_tab[rel_tab_ptr].symbol,symbol[0],MAX_SYM-1);
		  rel_tab[rel_tab_ptr].symbol[MAX_SYM-1]=0;
	    	  inc_rel_pointer();
	    	}
	    	else{
		  work_code |= disp[0];
	    	}
	    	break;
	      case INDIRECT:
	    	if(disp[0]>0x00FF){ 
	    	  sprintf(error,
		    "Displacement must fit into an unsigned 8-bit field");
	    	  return(-1);
	    	}
	    	work_code |= ((mod[0]<<11) | (reg_num[0]<<8) | disp[0]);
	    	break;
	      case IMMEDIATE:
	    	if(symbol[0][0]){ /* Symbol */
	    	  rel_tab[rel_tab_ptr].section=current_section;
	    	  rel_tab[rel_tab_ptr].offset=sec_loc_cnt[current_section].lc;
	    	  rel_tab[rel_tab_ptr].image_offset=image_count;
	    	  rel_tab[rel_tab_ptr].flags=R_RELWORD;
	    	  strncpy(rel_tab[rel_tab_ptr].symbol,symbol[0],MAX_SYM-1);
		  rel_tab[rel_tab_ptr].symbol[MAX_SYM-1]=0;
	    	  inc_rel_pointer();
	      	}
	      	else{
		  work_code |= disp[0];
	      	}
	      	break;
    	    }
	    break;

	case LDP:

	    if(parm_num<1 || parm_num>2 || type[0]!=IMMEDIATE
			|| (parm_num==2 && type[1]!=REGISTER) ){
		strcpy(error,
		 "Instruction requires expression and optional register");
		return(-1);
	    }
	    work_code = (type[0]<<21);
	    if(parm_num==2)
		work_code |= (reg_num[1]<<16);
	    else
		work_code |= (0x10L<<16);

    	    if(symbol[0][0]){ /* Symbol */
	    	rel_tab[rel_tab_ptr].section=current_section;
	    	rel_tab[rel_tab_ptr].offset=sec_loc_cnt[current_section].lc;
	    	rel_tab[rel_tab_ptr].image_offset=image_count;
	    	rel_tab[rel_tab_ptr].flags=R_RELPAGE;
	    	strncpy(rel_tab[rel_tab_ptr].symbol,symbol[0],MAX_SYM-1);
		rel_tab[rel_tab_ptr].symbol[MAX_SYM-1]=0;
	    	inc_rel_pointer();
	    }
	    else{
		work_code |= (disp[0]>>16);
    	    }
	    break;

	default:

	    sprintf(error,"Invalid special number(%d) in opcode table"
		,special_number);
	    return(-1);
	    break;
    }

    *parm_field=work_code;
    return(0);
}

long_immediate(parm_num,type,disp,symbol,error,parm_field)
int parm_num;
unsigned long type[],disp[],*parm_field;
char symbol[][MAX_SYM],*error;
{
    unsigned long work_code=0;

    *parm_field=0;

    if(parm_num != 1){
    	sprintf(error,"Instruction requires 1 operands");
    	return(-1);
    }
    switch((int)type[0]){	/* Switch on destination type */
	case REGISTER:
	case DIRECT:
	case INDIRECT:
	    sprintf(error,"parameter must be Immediate");
	    return(-1);
	    break;
	case IMMEDIATE:
	    if(symbol[0][0]){ /* Symbol */
	    	rel_tab[rel_tab_ptr].section=current_section;
	    	rel_tab[rel_tab_ptr].offset=sec_loc_cnt[current_section].lc;
	    	rel_tab[rel_tab_ptr].image_offset=image_count;
	    	rel_tab[rel_tab_ptr].flags=R_REL24;
	    	strncpy(rel_tab[rel_tab_ptr].symbol,symbol[0],MAX_SYM-1);
		rel_tab[rel_tab_ptr].symbol[MAX_SYM-1]=0;
	    	inc_rel_pointer();
	    }
	    else{
		 work_code |= disp[0];
	    }
	    break;
    }

    *parm_field=work_code;
    return(0);
}

trap_op(parm_num,type,disp,error,parm_field)
int parm_num;
unsigned long type[],disp[],*parm_field;
char *error;
{
    *parm_field=0;

    if(parm_num != 1){
    	sprintf(error,"Instruction requires 1 operand");
    	return(-1);
    }
    if(type[0] != IMMEDIATE){
    	sprintf(error,"Parameter must be immediate data");
    	return(-1);
    }
    if(disp[0] > 31){
    	sprintf(error,"Trap vector must be between 0 and 31 inclusive");
    	return(-1);
    }
    *parm_field = disp[0];
    return(0);
}

three_op(parm_num,type,reg_num,mod,disp,group,error,parm_field)
int parm_num,group;
unsigned long type[],reg_num[],mod[],disp[],*parm_field;
char *error;
{
    unsigned long work_code=0,t_field;
    int required_parms;

    *parm_field=0;
    required_parms=(group&FLAG)?2:3;	/* If flag set no destination */
    if(parm_num != required_parms){
    	sprintf(error,"Instruction requires %d operands",required_parms);
    	return(-1);
    }
    if(required_parms==3 && type[2] != REGISTER){
        sprintf(error,"Destination must be a register");
        return(-1);
    }
    if(required_parms==3 && (group & FLOAT_DEST) && reg_num[2]>7){ 
     	sprintf(error,"Destination requires extended precesion register");
        return(-1);
    }
	/* Determine T field */
    t_field = -1;
    if(type[1]==REGISTER && type[0]==REGISTER) t_field=0;
    if(type[1]==INDIRECT && type[0]==REGISTER) t_field=1;
    if(type[1]==REGISTER && type[0]==INDIRECT) t_field=2;
    if(type[1]==INDIRECT && type[0]==INDIRECT) t_field=3;
    if(t_field == -1){ /* Invalid combo */
	sprintf(error,
  	  "Only register and indirect sources permitted for 3 op instructions");
	return(-1);
    }
    work_code = (t_field<<21);
    if(required_parms==3) work_code |= (reg_num[2]<<16);
    switch((int)t_field){
	case 0: /* s1=register s2=register */
	    if((group & FLOAT_SOURCE) && reg_num[0]>7){ 
	    	sprintf(error,"Source 1 requires extended precision register");
	    	return(-1);
	    }
	    if((group & FLOAT_SOURCE) && reg_num[1]>7){ 
	    	sprintf(error,"Source 2 requires extended precesion register");
	    	return(-1);
	    }
	    work_code |= ( (reg_num[1]<<8) | reg_num[0]);
	    break;
	case 1: /* s1=indirect s2=register */
	    if((group & FLOAT_SOURCE) && reg_num[1]>7){ 
	    	sprintf(error,"Source 2 requires extended precesion register");
	    	return(-1);
	    }
	    if(mod[1]<8 && disp[1]>1){ 
	    	sprintf(error,"Disp must be 0,1,ir0,ir1 for 3 op instuctions");
	    	return(-1);
	    }
	    if(mod[1]<8 && disp[1]==0) /* Change addressing mode to no disp */
		mod[1]=0x18;
	    work_code |= ( (mod[1]<<11) | (reg_num[1]<<8) | reg_num[0]);
	    break;
	case 2: /* s1=register s2=indirect */
	    if((group & FLOAT_SOURCE) && reg_num[0]>7){ 
	    	sprintf(error,"Source 1 requires extended precesion register");
	    	return(-1);
	    }
	    if(mod[0]<8 && disp[0]>1){ 
	    	sprintf(error,"Disp must be 0,1,ir0,ir1 for 3 op instuctions");
	    	return(-1);
	    }
	    if(mod[0]<8 && disp[0]==0) /* Change addressing mode to no disp */
		mod[0]=0x18;
	    work_code |= ( (reg_num[1]<<8) | (mod[0]<<3) | reg_num[0] );
	    break;
	case 3: /* s1=indirect s2=indirect */
	    if(mod[0]<8 && disp[0]>1){ 
	    	sprintf(error,"Disp must be 0,1,ir0,ir1 for 3 op instuctions");
	    	return(-1);
	    }
	    if(mod[0]<8 && disp[0]==0) /* Change addressing mode to no disp */
		mod[0]=0x18;
	    if(mod[1]<8 && disp[1]>1){ 
	    	sprintf(error,"Disp must be 0,1,ir0,ir1 for 3 op instuctions");
	    	return(-1);
	    }
	    if(mod[1]<8 && disp[1]==0) /* Change addressing mode to no disp */
		mod[1]=0x18;
	    work_code |= ((mod[1]<<11)|(reg_num[1]<<8)|(mod[0]<<3)|reg_num[0]);
	    break;
    }
    *parm_field = work_code;
    return(0);
}

general_store(parm_num,type,reg_num,mod,disp,symbol,group,error,parm_field)
int parm_num,group;
unsigned long type[],reg_num[],mod[],disp[],*parm_field;
char symbol[][MAX_SYM],*error;
{
    unsigned long work_code=0;

    *parm_field=0;

    if(parm_num != 2){
    	sprintf(error,"Instruction requires 2 operands");
    	return(-1);
    }
    if(type[0] != REGISTER){
        sprintf(error,"Source must be a register");
        return(-1);
    }
    if((group & FLOAT_SOURCE) && reg_num[0]>7){ 
     	sprintf(error,"Source requires extended precesion register");
        return(-1);
    }
    work_code=(type[1]<<21) | (reg_num[0]<<16);
    switch((int)type[1]){	/* Switch on destination type */
	case REGISTER:
	    sprintf(error,"Destination must be Direct or Indirect");
	    return(-1);
	    break;
	case DIRECT:
	    if(symbol[1][0]){ /* Symbol */
	    	rel_tab[rel_tab_ptr].section=current_section;
	    	rel_tab[rel_tab_ptr].offset=sec_loc_cnt[current_section].lc;
	    	rel_tab[rel_tab_ptr].image_offset=image_count;
	    	rel_tab[rel_tab_ptr].flags=R_RELWORD;
	    	strncpy(rel_tab[rel_tab_ptr].symbol,symbol[1],MAX_SYM-1);
		rel_tab[rel_tab_ptr].symbol[MAX_SYM-1]=0;
	    	inc_rel_pointer();
	    }
	    else{
		work_code |= disp[1];
	    }
	    break;
	case INDIRECT:
	    if(disp[1]>0x00FF){ 
	    	sprintf(error,
		    "Displacement must fit into an unsigned 8-bit field");
	    	return(-1);
	    }
	    work_code |= ((mod[1]<<11) | (reg_num[1]<<8) | disp[1]);
	    break;
	case IMMEDIATE:
	    sprintf(error,"Destination must be Direct or Indirect");
	    return(-1);
	    break;
    }
    *parm_field = work_code;
    return(0);
}

general_load(parm_num,type,reg_num,mod,disp,symbol,group,error,parm_field)
int parm_num,group;
unsigned long type[],reg_num[],mod[],disp[],*parm_field;
char symbol[][MAX_SYM],*error;
{
    unsigned long work_code=0;

    *parm_field=0;

    if(parm_num != 2){
    	sprintf(error,"Instruction requires 2 operands");
    	return(-1);
    }
    if(type[1] != REGISTER){
        sprintf(error,"Destination must be a register");
        return(-1);
    }
    if((group & FLOAT_DEST) && reg_num[1]>7){ 
     	sprintf(error,"Destination requires extended precesion register");
        return(-1);
    }
    work_code=(type[0]<<21) | (reg_num[1]<<16);
    switch((int)type[0]){
	case REGISTER:
	    if(group & INTERLOCKED){ /* Invalid source */
	    	sprintf(error,"Source must be Direct or Indirect");
	    	return(-1);
	    }
	    if((group & FLOAT_SOURCE) && reg_num[0]>7){ 
	    	sprintf(error,"Source requires extended precesion reg");
	    	return(-1);
	    }
	    work_code |= reg_num[0];
	    break;
	case DIRECT:
	    if(symbol[0][0]){ /* Symbol */
	    	rel_tab[rel_tab_ptr].section=current_section;
	    	rel_tab[rel_tab_ptr].offset=sec_loc_cnt[current_section].lc;
	    	rel_tab[rel_tab_ptr].image_offset=image_count;
	    	rel_tab[rel_tab_ptr].flags=R_RELWORD;
	    	strncpy(rel_tab[rel_tab_ptr].symbol,symbol[0],MAX_SYM-1);
		rel_tab[rel_tab_ptr].symbol[MAX_SYM-1]=0;
	    	inc_rel_pointer();
	    }
	    else{
		work_code |= disp[0];
	    }
	    break;
	case INDIRECT:
	    if(disp[0]>0x00FF){ 
	    	sprintf(error,
		    "Displacement must fit into an unsigned 8-bit field");
	    	return(-1);
	    }
	    work_code |= ((mod[0]<<11) | (reg_num[0]<<8) | disp[0]);
	    break;
	case IMMEDIATE:
	    if(symbol[0][0]){ /* Symbol */
	    	rel_tab[rel_tab_ptr].section=current_section;
	    	rel_tab[rel_tab_ptr].offset=sec_loc_cnt[current_section].lc;
	    	rel_tab[rel_tab_ptr].image_offset=image_count;
	    	rel_tab[rel_tab_ptr].flags=R_RELWORD;
	    	strncpy(rel_tab[rel_tab_ptr].symbol,symbol[0],MAX_SYM-1);
		rel_tab[rel_tab_ptr].symbol[MAX_SYM-1]=0;
	    	inc_rel_pointer();
	    }
	    else{
		 work_code |= disp[0];
	    }
	    break;
    }
    *parm_field = work_code;
    return(0);
}

static char index_add[][15]={
		"*+ar?(^)","*-ar?(^)","*++ar?(^)","*--ar?(^)",
		"*ar?++(^)","*ar?--(^)","*ar?++(^)%","*ar?--(^)%",
		"*+ar?(ir0)","*-ar?(ir0)","*++ar?(ir0)","*--ar?(ir0)",
		"*ar?++(ir0)","*ar?--(ir0)","*ar?++(ir0)%","*ar?--(ir0)%",
		"*+ar?(ir1)","*-ar?(ir1)","*++ar?(ir1)","*--ar?(ir1)",
		"*ar?++(ir1)","*ar?--(ir1)","*ar?++(ir1)%","*ar?--(ir1)%",
		"*ar?","*ar?++(ir0)b"
			    };

char c30_regs[][4]={"r0","r1","r2","r3","r4","r5","r6","r7"
		   ,"ar0","ar1","ar2","ar3","ar4","ar5","ar6","ar7"
		   ,"dp","ir0","ir1","bk","sp","st","ie","if","iof"
		   ,"rs","re","rc"};

/*****************************************
**					**
**	type=0 Register			**
**	     1 Direct			**
**	     2 Indirect			**
**	     3 Immediate		**
**					**
*****************************************/

decode_addr_mode(disp_24,itype,expr,type,reg_num,mod,disp,symbol,error)
int disp_24,itype;	/* itype=0 must be int.  itype=1 must be float */
char *expr,*symbol,*error;
unsigned long *type,*reg_num,*mod;
unsigned long *disp;
{
    int i,tmp_reg,index,found;
    char c,string[100],*ptr;
    unsigned long mod_field;

    *type=0L;
    *reg_num=0L;
    *mod=0L;
    *disp=0L;
    *symbol=0;

    if(expr[0] == '@') *type=1;  /* Direct */
    else{
 	if(expr[0] == '*') *type=2;	 /* Indirect */
        else{	/* register or immediate */
	    if(reg_mode(expr,reg_num)){
		*type=0L; 
		return(0);
	    }
	    else{  /* Immediate */
		*type=3;
	    }
	}
    }
     switch((int)*type){
	case 0: /* Register */
		break;
	case 1: /* Direct Addressing */
   		if(expr[1]>='0' && expr[1]<='9'){/* Constant */
	    	    if(get_const(0,2,&expr[1],disp)){
	    	    	sprintf(error,"Invalid Constant");
	            	return(-1);
	    	    }
	     	    symbol[0]=0;
		}
		else{ /* Symbol */
	    	    strcpy(symbol,&expr[1]);
	    	    *disp=0L;
		}
		break;
	case 2: /* Indirect */
		for(mod_field=NUM_IND_MODES-1;mod_field>=0;mod_field--){
		    string[0]=0;
		    for(i=index=0;i<15;i++){
			if(mod_field < 6){ /* check for shorthand addressing */
			  if(expr[index]==0 && index_add[mod_field][i]=='('){
			    *reg_num=tmp_reg;
			    *mod=mod_field;
			    *disp=1L;
			    return(0);
			  }
			}
			if(index_add[mod_field][i]==expr[index]){
			    if(expr[index]==0){  /* Match */
				*reg_num=tmp_reg;
				*mod=mod_field;
				c=string[0];
				if(c){
    				    if(c>='0' && c<='9'){ /* Constant */
				   	if(get_const(0,4,string,disp)){
  					    sprintf(error,
			       "Invalid indirect addressing displacement \"%s\""
					    	,string);
				   	    return(-1);
				    	}
				    }
				    else{ /* symbol */
		   			for(i=found=0;i<sym_ptr;i++){
					    if(!strcmp(sym_tab[i].symbol
								      ,string)){ 
                	    			found=1;
                	    			break;
            				    }
        	    			}
		    			if(!found){
                        		    sprintf(error,
			    	 "symbol \"%s\" not previously defined",string);
                    			    return(-1);
		    			}
		    			*disp=sym_tab[i].add;
				    }
				}
				return(0);
			    }
			    index++;
			}
			else{
			    if(index_add[mod_field][i]=='?'){
			    	if(expr[index]<'0' || expr[index]>'7'){
				    sprintf(error,"Invalid register");
				    return(-1);
			    	}
			    	tmp_reg=expr[index]-'0';
			    	index++;
			    }
			    else{
			        if(index_add[mod_field][i]=='^'){
				    ptr=string;
				    while(expr[index] && expr[index]!=')'){
					*ptr++ = expr[index++];
				    }
				    *ptr=0;
				}
				else{
				    i=1000;  /* no match */
				}
			    }
			}
		    }
		}
	        sprintf(error,"Invalid indirect addressing mode \"%s\"",expr);
	        return(-1);
		break;
	case 3: /*Immediate */
   		if((expr[0]>='0' && expr[0]<='9')||expr[0]=='-'){ /* Constant */
	     	    symbol[0]=0;
	    	    if(get_const(1,(disp_24)?3:2,&expr[0],disp)){
	    	    	sprintf(error,"Invalid Constant");
	            	return(-1);
	    	    }
		    if(inquire_last_const() != itype){
			strcpy(error,"Constant must be ");
			if(itype)
			    strcat(error,"float");
			else
			    strcat(error,"integer");
			return(-1);
		    }
		}
		else{ /* Symbol */
	    	    strcpy(symbol,&expr[0]);
	    	    *disp=0L;
		}
		break;
	default:
		sprintf(error,"Invalid parameter \"%s\"",expr);
		return(-1);
		break;
    }
    return(0);
}

reg_mode(expr,reg_num)
char *expr;
unsigned long *reg_num;
{
    register int i;
    
    for(i=0;i<NUM_REGS;i++){
    	if(!strcmp(c30_regs[i],expr)){
	    *reg_num=(unsigned long)i;
	    return(1);
	}
    }
    *reg_num=0L;
    return(0);	/* not register mode */
}

eval_line(parameters,line_num,error,type,num_bytes)
char *parameters,*error;
int line_num,type,num_bytes;
{
    char expr[150],*ptr=expr,*ptr2;
    int i;

    for( ; *parameters; parameters++){
	if(*parameters != ','){
	    *ptr++ = *parameters;
    	}
    	else{
	    *ptr=0;
	    ptr=expr;
	    for(i=0;isspace(expr[i]) && expr[i];i++); /* Remove leading spaces*/
	    ptr2=&expr[i];
	    while(*ptr2){	/* Remove trailing space */
		if(isspace(*ptr2)) *ptr2=0;
		ptr2++;
	    }
	    if(eval_expression(&expr[i],line_num,error,type,num_bytes)) 
		return(-1);
    	}
    }
    *ptr=0;
    if(expr[0]){
	for(i=0;isspace(expr[i]) && expr[i];i++); /* Remove leading spaces*/
	ptr2=&expr[i];
	while(*ptr2){	/* Remove trailing space */
	    if(isspace(*ptr2)) *ptr2=0;
	    ptr2++;
	}
        if(eval_expression(&expr[i],line_num,error,type,num_bytes)) return(-1);
    }
    else{
        img_ptr->address=sec_loc_cnt[current_section].lc;
	img_ptr->file_number=fp_stack_pointer;
    	img_ptr->line_num=line_num;  
    	img_ptr->section=current_section;  
    	img_ptr->data=0L;
    	inc_image_pointer();
    	sec_loc_cnt[current_section].lc++;
	sprintf(error,"Expression or symbol required");
	return(-1);
    }
    return(0);
}

/*  
    type=0 .word
    type=1 .float
*/

eval_expression(expr,line_num,error,type,num_bytes)
char *expr,*error;
int line_num,type,num_bytes;
{
    unsigned long word_value;
    int return_val=0;
    float f;
    double atof();

    if( (expr[0]>='0' && expr[0]<='9') || expr[0]=='-'){	/* Constant */
	switch(type){
	    case 0: /* .word */
		return_val=get_const(0,num_bytes,expr,&word_value);
		break;
	    case 1: /* .float */
   		f=(float)atof(expr);
    		return_val=convert_TI_float(LONG,f,&word_value);
		break;
	}
	if(return_val){
	    if(return_val == -1)
		sprintf(error,"Invalid Constant \"%s\"",expr);
	    else
		sprintf(error,"Constant too large");
	 }
    }
    else{	/* Symbol, Charactor or String */
	if(expr[0]=='\''){ /* Charactor */
	    word_value=(unsigned long)expr[1];
	}
	else if(expr[0]=='"'){ /* String */
	    word_value=0L;
	}
	else{ /* Symbol */
	    rel_tab[rel_tab_ptr].section=current_section;
	    rel_tab[rel_tab_ptr].offset=sec_loc_cnt[current_section].lc;
	    rel_tab[rel_tab_ptr].image_offset=image_count;
	    rel_tab[rel_tab_ptr].flags=R_RELLONG;
#ifdef MAC
	    strncpy(rel_tab[rel_tab_ptr].symbol,expr,(size_t)(MAX_SYM-1));
#else
	    strncpy(rel_tab[rel_tab_ptr].symbol,expr,MAX_SYM-1);
#endif
	    rel_tab[rel_tab_ptr].symbol[MAX_SYM-1]=0;
	    inc_rel_pointer();
            word_value=0L;
	}
    }
    img_ptr->address=sec_loc_cnt[current_section].lc;
    img_ptr->file_number=fp_stack_pointer;
    img_ptr->line_num=line_num;  
    img_ptr->section=current_section;  
    img_ptr->data=word_value;
    inc_image_pointer();
    sec_loc_cnt[current_section].lc++;
    return(return_val);
}

static int last_const=0;

get_const(en_float,num_bytes,string,value)
int en_float,num_bytes;
char string[];
unsigned long *value;
{
    char c,*ptr=string;
    int i,len,type,negative=0;
    unsigned long work,mask;
    float f;

    while(*ptr){       /* Remove trailing space */
	if(isspace(*ptr)) *ptr=0;
	ptr++;
    }
    ptr=string;
    work=0L;  type=2;
    switch(num_bytes){
	case 1:
		mask=0xFFFFFF00L;  break;
	case 2:
		mask=0xFFFF0000L;  break;
	case 3:
		mask=0xFFFF0000L;  break;
	case 4:
		mask=0x00000000L;  break;
    }
    for(len=0;*ptr;len++){
	if(*ptr++ =='.') type=4;
    }
    if(!en_float) type=2;
    if(type==2){
	if(string[len-1]=='B' || string[len-1]=='b')
	    {type=0; string[len-1]=0; len--;} 
	else if(string[len-1]=='Q' || string[len-1]=='q')
	    {type=1; string[len-1]=0; len--;} 
	else if(string[len-1]=='H' || string[len-1]=='h')
	    {type=3; string[len-1]=0; len--;} 
	else if(string[0]=='0' && string[1]=='x')
	    {type=3; string+=2; len-=2; }
    }
    *value=0L;
    switch(type){
	case 0: /* Binary */
		last_const=0;
		if( len > (num_bytes<<3) ) return(-2);
		for(i=0;i<len;i++){
		    if(string[i]!='0' && string[i]!='1') return(-1);
		    work=(work<<1)+(string[i]-'0');
		}
		*value=work;
		break;
	case 1: /* Octal */
		if(len>11) return(-2);
		for(i=0;i<len;i++){
		    c=string[i];
		    if(c<'0' || c>'7') return(-1);
		    work=(work<<3)+(string[i]-'0');
		}
		if( (work&mask) != 0L) return(-2); /* Value too large */
		*value=work;
		break;
	case 2: /* Decimal */
		last_const=0;
		if(string[0]=='-'){
		    negative=1; len--; string++;
		}
		if(len>10) return(-2);
		for(i=0;i<len;i++){
		    c=string[i];
		    if(c<'0' || c>'9') return(-1);
		    work=( (work<<3) + (work<<1) ) + c-'0';
		}
		if( (work&mask) != 0L) return(-2); /* Value too large */
		*value=(negative)?(-work):work;
		*value &= (~mask);
		break;
	case 3: /* Hex */
		last_const=0;
		if(len>1 && string[0]=='0'){len--; string++; }
		if(len>(num_bytes+num_bytes)) return(-2);
		for(i=0;i<len;i++){
		    c=string[i]; work=work<<4;
		    if(c>='0' && c<='9') work+=(c-'0');
	  	    else{
			c=toupper(c);
			if(c>='A' && c<='F') work+=(c-'A'+10);
			else return(-1);
		    }
		}
		*value=work;
		break;
	case 4:	/* float */
		last_const=1;
   		f=atof(string);
    		return(convert_TI_float(SHORT,f,value));
		break;
    }
    return(0);
}

inquire_last_const()
{
    return(last_const);
}
